@@ -67,6 +67,7 @@ typedef enum : unsigned {
         VTE_DEBUG_BIDI          = 1u << 28,
         VTE_DEBUG_CONVERSION    = 1u << 29,
         VTE_DEBUG_EXCEPTIONS    = 1u << 30,
+        VTE_DEBUG_IMAGE         = 1u << 31,
 } VteDebugFlags;
 
 void _vte_debug_init(void);
@@ -112,6 +113,12 @@ static void _vte_debug_print(guint flags, const char *fmt, ...)
 #define _vte_debug_print(args...) do { } while(0)
 #endif /* VTE_DEBUG */
 
+static inline char const*
+_vte_debug_tf(bool v) noexcept
+{
+        return v ? "true" : "false";
+}
+
 G_END_DECLS
 
 #endif
