@@ -2,19 +2,18 @@
  * Copyright (C) 2001,2002 Red Hat, Inc.
  * Copyright © 2009, 2010, 2019 Christian Persch
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 3 of the License, or (at your option) any later version.
+ * This library is free software: you can redistribute it and/or modify
+ * it under the terms of the GNU Lesser General Public License as published
+ * by the Free Software Foundation, either version 3 of the License, or
+ * (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this library.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 /**
@@ -39,9 +38,6 @@
 #include <sys/types.h>
 #include <sys/ioctl.h>
 #include <sys/socket.h>
-#ifdef HAVE_SYS_TERMIOS_H
-#include <sys/termios.h>
-#endif
 #include <errno.h>
 #include <fcntl.h>
 #include <limits.h>
@@ -53,9 +49,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#ifdef HAVE_TERMIOS_H
 #include <termios.h>
-#endif
 #include <unistd.h>
 #ifdef HAVE_UTIL_H
 #include <util.h>
@@ -294,8 +288,10 @@ Pty::set_size(int rows,
 	memset(&size, 0, sizeof(size));
 	size.ws_row = rows > 0 ? rows : 24;
 	size.ws_col = columns > 0 ? columns : 80;
+#ifdef WITH_SIXEL
         size.ws_ypixel = size.ws_row * cell_height_px;
         size.ws_xpixel = size.ws_col * cell_width_px;
+#endif
 	_vte_debug_print(VTE_DEBUG_PTY,
 			"Setting size on fd %d to (%d,%d).\n",
 			master, columns, rows);
@@ -355,10 +351,18 @@ Pty::get_size(int* rows,
 static int
 fd_set_cpkt(vte::libc::FD& fd)
 {
+        auto ret = 0;
+#if defined(TIOCPKT)
         /* tty_ioctl(4) -> every read() gives an extra byte at the beginning
          * notifying us of stop/start (^S/^Q) events. */
         int one = 1;
-        return ioctl(fd.get(), TIOCPKT, &one);
+        ret = ioctl(fd.get(), TIOCPKT, &one);
+#elif defined(__sun) && defined(HAVE_STROPTS_H)
+        if (isastream(fd.get()) == 1 &&
+            ioctl(fd.get(), I_FIND, "pckt") == 0)
+                ret = ioctl(fd.get(), I_PUSH, "pckt");
+#endif
+        return ret;
 }
 
 static int
@@ -398,7 +402,7 @@ fd_setup(vte::libc::FD& fd)
                 auto errsv = vte::libc::ErrnoSaver{};
                 _vte_debug_print(VTE_DEBUG_PTY,
                                  "%s failed: %s",
-                                 "ioctl(TIOCPKT)", g_strerror(errsv));
+                                 "Setting packet mode", g_strerror(errsv));
                 return -1;
         }
 
@@ -464,7 +468,7 @@ _vte_pty_open_posix(void)
                 auto errsv = vte::libc::ErrnoSaver{};
                 _vte_debug_print(VTE_DEBUG_PTY,
                                  "%s failed: %s",
-                                 "ioctl(TIOCPKT)", g_strerror(errsv));
+                                 "Setting packet mode", g_strerror(errsv));
                 return {};
         }
 
